import functools
import inspect
import sys


class _StdoutHandler:
    def __init__(self, filename):
        self.filename = filename

    def write(self, text):
        with open(self.filename, 'a') as f:
            f.write(text)

    def flush(self):
        pass


class _Print2File:
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        original_stdout = sys.stdout
        sys.stdout = _StdoutHandler(f'{inspect.stack()[2].filename.replace(".py", "")}-{self.func.__name__}.txt')
        try:
            result = self.func(*args, **kwargs)
            return result
        finally:
            sys.stdout = original_stdout

    def __get__(self, instance, owner):
        return functools.partial(self.__call__, instance)


def print2file(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return _Print2File(func)(*args, **kwargs)

    return wrapper



from __future__ import annotations
from typing import Union


def _contain(l1: list[tuple], l2: list[tuple]) -> bool:
    if l1 is None:
        return False

    for i in l2:

        if i[1] is None:
            if i[0] not in [j[0] for j in l1]:
                return False

        elif i not in l1:
            return False

    return True


class Element:
    __slots__ = ('tag', 'args', 'data', 'text', 'parent', 'children', '_arg_dict')

    def __init__(
            self,
            tag: str,
            args: list[tuple[str, str]] = None,
            parent: Element = None,
    ):
        self.tag = tag
        self.args = args
        self.data = ''
        self.text = ''
        self.parent = parent
        self.children = []

        self._arg_dict = dict()
        if args is None:
            return
        for i in args:
            self._arg_dict[i[0]] = i[1]

    # Depth first search
    def find(self, tag: str = None, args: dict = None) -> Union[Element, None]:

        if tag is None and args is None:
            raise AttributeError("tag and args can't both be None!")

        # if tag is handed, pair it firstly
        if tag is not None:

            if self.tag == tag:
                if args is not None:
                    args_items = list(args.items())  # dict -> list[tuple]
                    if _contain(self.args, args_items):
                        return self
                else:
                    return self

            # leaf node
            if len(self.children) == 0:
                return None

            # recurse
            for i in self.children:
                res = i.find(tag, args)
                if res is not None:
                    return res

        # only args is handed
        else:

            args_items = list(args.items())  # dict -> list[tuple]
            if _contain(self.args, args_items):
                return self

            # leaf node
            if len(self.children) == 0:
                return None

            # recurse
            for i in self.children:
                res = i.find(tag, args)
                if res is not None:
                    return res

        return None

    def find_all(self, tag: str = None, args: dict = None) -> list[Element]:

        if tag is None and args is None:
            raise AttributeError("tag and args can't both be None!")

        res = []

        if tag is not None:

            if self.tag == tag:
                if args is not None:
                    args_items = list(args.items())  # dict -> list[tuple]
                    if _contain(self.args, args_items):
                        res.append(self)
                else:
                    res.append(self)

            # leaf node
            if len(self.children) == 0:
                return res

            # recurse
            for i in self.children:
                res.extend(i.find_all(tag, args))

        # only args are handed
        elif args is not None:

            args_items = list(args.items())  # dict -> list[tuple]
            if _contain(self.args, args_items):
                res.append(self)

            # leaf node
            if len(self.children) == 0:
                return res

            # recurse
            for i in self.children:
                res.extend(i.find_all(tag, args))

        else:
            return [self]

        return res

    def copy(self):
        node = Element(self.tag, self.args, self.parent)
        node.data = self.data
        node.text = self.text
        node.children = self.children.copy()
        return node

    def __repr__(self):
        return f'<Element {self.tag}, {self.args}>'

    def __getitem__(self, item):
        if type(item) is str:
            return self._arg_dict[item]
        elif type(item) is int:
            return self.children[item]
        else:
            raise AttributeError('str or int needed!')
           
    def __iter__(self):
        return iter(self.children)        
    



import os
import ctypes
import sys
import subprocess


def add_context():
    """
    Add current file's directory into system path.
    """
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))


def get_admin():
    """run script with administrator privilege
    """
    if ctypes.windll.shell32.IsUserAnAdmin():
        pass
    else:
        # rerun the script with admin privilege
        ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, " ".join(sys.argv), None, 1)


def exec_ps(ps_path: str, show_result=False, show_error=True):
    """Execute PowerShell script line by line

    Args:
        ps_path (str): path to PowerShell script
        show_result (bool, optional): whether to show the result or not. Defaults to False.
        show_error (bool, optional): whether to show the error or not. Defaults to True.
    """
    stdout = sys.stdout if show_result == True else None
    stderr = sys.stdout if show_error == True else None
    shell = subprocess.Popen(
        ["powershell.exe"],
        stdin=subprocess.PIPE,
        stdout=stdout,
        stderr=stderr,
        shell=True,
        universal_newlines=True
    )
    with open(ps_path, 'r') as f:
        for i in f.readlines():
            shell.stdin.write(i)
            shell.stdin.flush()



import logging

_LEVEL = logging.DEBUG
_FORMAT = '[%(name)s] - %(levelname)s - %(message)s'


class Logger(logging.Logger):
    def __init__(self, name: str):
        super().__init__(name)
        formatter = logging.Formatter(fmt=_FORMAT)
        handler = logging.StreamHandler()
        handler.setLevel(_LEVEL)
        handler.setFormatter(formatter)
        self.addHandler(handler)
        self.setLevel(_LEVEL)
